﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using System.Xml;
using Fiddler;

namespace IEGeolocMock
{
    public class LocationState
    {
        private bool running = false;
        private string latitude = "47.6730";
        private string longitutde = "-122.1154";
        private string altitude = "0";
        private string uncertainty = "0";
        private ReaderWriterLock stateLock = new ReaderWriterLock();
        private int processId = System.Diagnostics.Process.GetCurrentProcess().Id;

        public LocationState()
        {
            StartFiddler();
        }

        public void Stop()
        {
            StopFiddler();
        }

        public void Get(out string latitude, out string longitutde, out string altitude, out string uncertainty)
        {
            try
            {
                this.stateLock.AcquireReaderLock(1000);
                latitude = this.latitude;
                longitutde = this.longitutde;
                altitude = this.altitude;
                uncertainty = this.uncertainty;
            }
            finally
            {
                this.stateLock.ReleaseReaderLock();
            }
        }

        public delegate void LocationUpdatedHandler(object sender, EventArgs args);
        public event LocationUpdatedHandler LocationUpdated;

        private bool IsDouble(string intString)
        {
            double result = 0;
            return Double.TryParse(intString, out result);
        }

        public void Set(string latitude, string longitutde, string altitude, string uncertainty)
        {
            try
            {
                this.stateLock.AcquireWriterLock(1000);

                if (IsDouble(latitude))
                {
                    this.latitude = latitude;
                }

                if (IsDouble(longitutde))
                {
                    this.longitutde = longitutde;
                }

                if (IsDouble(altitude))
                {
                    this.altitude = altitude;
                }

                if (IsDouble(uncertainty))
                {
                    this.uncertainty = uncertainty;
                }

                if (this.LocationUpdated != null)
                {
                    this.LocationUpdated(this, null);
                }
            }
            finally
            {
                this.stateLock.ReleaseWriterLock();
            }

            // After updating values, clear out IE's cache so they'll ask us next time.
            ClearIEGeoCache();
        }

        public void ClearIEGeoCache()
        {
            Microsoft.Win32.Registry.CurrentUser.DeleteSubKeyTree("Software\\Microsoft\\Internet Explorer\\Geolocation\\CacheRoot", false);
        }

        private void StartFiddler()
        {
            if (!this.running)
            {
                this.running = true;

                if (!Fiddler.CertMaker.rootCertExists())
                {
                    if (!Fiddler.CertMaker.createRootCert())
                    {
                        throw new Exception("Unable to create cert for FiddlerCore.");
                    }
                }

                if (!Fiddler.CertMaker.rootCertIsTrusted())
                {
                    if (!Fiddler.CertMaker.trustRootCert())
                    {
                        throw new Exception("Unable to install FiddlerCore's cert.");
                    }
                }

                Fiddler.FiddlerApplication.Startup(8282, 
                    FiddlerCoreStartupFlags.Default | 
                    FiddlerCoreStartupFlags.AllowRemoteClients |
                    FiddlerCoreStartupFlags.DecryptSSL | 
                    FiddlerCoreStartupFlags.MonitorAllConnections | 
                    FiddlerCoreStartupFlags.RegisterAsSystemProxy | 
                    FiddlerCoreStartupFlags.ChainToUpstreamGateway |
                    FiddlerCoreStartupFlags.CaptureLocalhostTraffic);

                Fiddler.FiddlerApplication.BeforeRequest += new SessionStateHandler(FiddlerApplication_BeforeRequest);
                Fiddler.FiddlerApplication.BeforeResponse += new SessionStateHandler(FiddlerApplication_BeforeResponse);
            }
        }

        private void FiddlerApplication_BeforeRequest(Session oSession)
        {
            if (IsInterestingSession(oSession))
            {
                oSession.bBufferResponse = true;
            }
        }

        private void FiddlerApplication_BeforeResponse(Session oSession)
        {
            if (IsInterestingSession(oSession))
            {
                if (!tryUpdateXml(oSession))
                {
                    tryUpdateJSON(oSession);
                }
            }
        }

        private void StopFiddler()
        {
            if (this.running)
            {
                Fiddler.FiddlerApplication.Shutdown();
                this.running = false;
            }
        }

        static private Uri googleLocUri = new Uri("https://www.google.com/loc/json");

        private bool IsInterestingSession(Session oSession)
        {
            Uri uri = new Uri(oSession.fullUrl);

            // We want to get the correct information for our process.  Only mock other processes.
            return oSession.LocalProcessID != System.Diagnostics.Process.GetCurrentProcess().Id
                && (uri.Host == "inference.location.live.net" || uri.Equals(googleLocUri));
        }

        private bool tryUpdateXml(Session oSession)
        {
            Uri uri = new Uri(oSession.fullUrl);
            // Example
            //      From: https://inference.location.live.net/inferenceservice/v21/Pox/GetLocationUsingFingerprint
            //      Body: // <?xml version="1.0" encoding="utf-8"?><GetLocationUsingFingerprintResponse xmlns="http://inference.location.live.com" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xsd="http://www.w3.org/2001/XMLSchema"><GetLocationUsingFingerprintResult><ResponseStatus>Success</ResponseStatus><LocationResult><ResolverStatus Status="Success" Source="Internal"/><ResolvedPosition Latitude="47.682478" Longitude="-122.110999" Altitude="0"/><RadialUncertainty>350</RadialUncertainty><TileResult/><TrackingId>8d7c3800-1cac-4c94-a82a-4c0a82ef6d0c</TrackingId></LocationResult><ExtendedV21Result CrowdSourcingLevel="Medium" ServerUtcTime="2011-04-02T03:42:56.8887515Z"/></GetLocationUsingFingerprintResult></GetLocationUsingFingerprintResponse>
            bool xmlDocModified = false;
            XmlDocument xmlDoc = new XmlDocument();
            try
            {
                oSession.utilDecodeResponse();
                xmlDoc.LoadXml(oSession.GetResponseBodyAsString());
            }
            catch (XmlException)
            {
                System.Diagnostics.Debug.WriteLine("Found URI <" + uri.ToString() + "> but unable to parse response as XML: " + oSession.GetResponseBodyAsString());
            }

            XmlNamespaceManager nsmgr = new XmlNamespaceManager(xmlDoc.NameTable);
            nsmgr.AddNamespace("map", "http://inference.location.live.com");

            try
            {
                this.stateLock.AcquireReaderLock(1000);
                var posNode = xmlDoc.SelectSingleNode("//map:ResolvedPosition", nsmgr);
                if (posNode != null)
                {
                    posNode.Attributes["Latitude"].Value = this.latitude;
                    posNode.Attributes["Longitude"].Value = this.longitutde;
                    posNode.Attributes["Altitude"].Value = this.altitude;
                    xmlDocModified = true;
                }

                var uncNode = xmlDoc.SelectSingleNode("//map:RadialUncertainty", nsmgr);
                if (uncNode != null)
                {
                    uncNode.InnerText = this.uncertainty;
                    xmlDocModified = true;
                }
            }
            finally
            {
                this.stateLock.ReleaseReaderLock();
            }

            if (xmlDocModified)
            {
                string newXmlStr = xmlDoc.OuterXml;
                System.Diagnostics.Debug.WriteLine("Found URI <" + uri.ToString() + "> and changed XML to: " + newXmlStr);
                oSession.utilSetResponseBody(newXmlStr);
                System.Diagnostics.Debug.WriteLine("Actually: " + oSession.GetResponseBodyAsString());
            }
            else
            {
                System.Diagnostics.Debug.WriteLine("Found URI <" + uri.ToString() + "> but didn't modify XML: " + xmlDoc.OuterXml);
            }

            return xmlDocModified;
        }

        private bool tryUpdateJSON(Session oSession)
        {
            /* Example Session to <https://www.google.com/loc/json>
            POST /loc/json HTTP/1.1
            Host: www.google.com
            Connection: keep-alive
            Content-Length: 513
            Pragma: no-cache
            Cache-Control: no-cache
            Content-Type: application/json
            User-Agent: Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/534.16 (KHTML, like Gecko) Chrome/10.0.648.204 Safari/534.16
            Accept-Encoding: gzip,deflate,sdch
            Accept-Language: en-US,en;q=0.8
            Accept-Charset: ISO-8859-1,utf-8;q=0.7,*;q=0.3

            {"host":"deletethis.net","radio_type":"unknown","request_address":false,"version":"1.1.0","wifi_towers":[{"age":0,"mac_address":"00-18-f8-70-97-9a","signal_strength":-52,"ssid":"sarah"},{"age":0,"mac_address":"00-21-63-21-fd-71","signal_strength":-91,"ssid":"PGQQX"},{"age":0,"mac_address":"00-21-63-22-13-5a","signal_strength":-89,"ssid":"hiroshi"},{"age":0,"mac_address":"00-23-08-10-7b-f5","signal_strength":-94,"ssid":"VYVX8"},{"age":0,"mac_address":"00-24-d2-11-0e-39","signal_strength":-91,"ssid":"AD5BK"}]}
            HTTP/1.1 200 OK
            Content-Type: application/json; charset=UTF-8
            Date: Sun, 03 Apr 2011 06:11:23 GMT
            Expires: Sun, 03 Apr 2011 06:11:23 GMT
            Cache-Control: private, max-age=0
            X-Content-Type-Options: nosniff
            X-Frame-Options: SAMEORIGIN
            X-XSS-Protection: 1; mode=block
            Content-Length: 129
            Server: GSE

            {"location":{"latitude":47.6828212,"longitude":-122.110555,"accuracy":42.0},"access_token":"2:3CWWcYZGssd9KZOj:q8_v6VC6_ljEjFgp"}
             */ 
            bool updated = false;

            if (oSession.oResponse.MIMEType.ToLowerInvariant() == "application/json")
            {
                string original = oSession.GetResponseBodyAsString();
                string modified = original;

                try
                {
                    this.stateLock.AcquireReaderLock(1000);

                    // Here's where I appologize for how I'm about to modify this JSON. In this version of .Net, there's no 
                    // standard JSON parser. So I can either, write one myself, hook up to a WebBrowser control and have it 
                    // do JSON parsing, or I can make assumptions about how this particular JSON is formatted.  Two of those
                    // options are a lot more work than the third...
                    modified = Regex.Replace(modified, "\"latitude\":[^,]*,", "\"latitude\":" + this.latitude + ",");
                    modified = Regex.Replace(modified, "\"longitude\":[^,]*,", "\"longitude\":" + this.longitutde + ",");
                    modified = Regex.Replace(modified, "\"accuracy\":[^\\}]*\\}", "\"accuracy\":" + this.uncertainty + "}");

                }
                finally
                {
                    this.stateLock.ReleaseReaderLock();
                }

                if (modified != original)
                {
                    updated = true;
                    oSession.utilSetResponseBody(modified);
                }
            }

            return updated;
        }
    }
}